(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('log4js'), require('node:fs/promises'), require('node:path'), require('fast-xml-parser'), require('node:fs'), require('node:crypto')) :
    typeof define === 'function' && define.amd ? define(['exports', 'log4js', 'node:fs/promises', 'node:path', 'fast-xml-parser', 'node:fs', 'node:crypto'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.jsTemplate = {}, global.log4js, global.promises, global.path, global.fastXmlParser, global.fs, global.node_crypto));
})(this, (function (exports, log4js, promises, path, fastXmlParser, fs, node_crypto) { 'use strict';

    function _interopNamespaceDefault(e) {
        var n = Object.create(null);
        if (e) {
            Object.keys(e).forEach(function (k) {
                if (k !== 'default') {
                    var d = Object.getOwnPropertyDescriptor(e, k);
                    Object.defineProperty(n, k, d.get ? d : {
                        enumerable: true,
                        get: function () { return e[k]; }
                    });
                }
            });
        }
        n.default = e;
        return Object.freeze(n);
    }

    var log4js__namespace = /*#__PURE__*/_interopNamespaceDefault(log4js);
    var promises__namespace = /*#__PURE__*/_interopNamespaceDefault(promises);
    var path__namespace = /*#__PURE__*/_interopNamespaceDefault(path);
    var fs__namespace = /*#__PURE__*/_interopNamespaceDefault(fs);

    /******************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    /* global Reflect, Promise, SuppressedError, Symbol */


    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }

    typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
        var e = new Error(message);
        return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
    };

    const XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>';
    const INDENT = "    ";

    const XML_PARSER_OPTION = {
        commentPropName: "!---",
        ignoreAttributes: false,
        ignoreNameSpace: false,
        parseTagValue: false,
        parseNodeValue: false,
        parseAttributeValue: false,
        trimValues: true,
        processEntities: false,
        cdataPropName: "![CDATA[",
    };
    const JSON_PARSER_OPTION = Object.assign(Object.assign({}, XML_PARSER_OPTION), { format: true, indentBy: INDENT, suppressBooleanAttributes: false, suppressEmptyNode: false });

    function buildReassembledFile(combinedXmlContents, filePath, xmlElement, xmlRootElementHeader) {
        return __awaiter(this, void 0, void 0, function* () {
            let finalXmlContent = combinedXmlContents.join("\n");
            finalXmlContent = finalXmlContent.replace(/<\?xml version="1.0" encoding="UTF-8"\?>/g, "");
            finalXmlContent = finalXmlContent.replace(new RegExp(`<${xmlElement}\\s*[^>]*>`, "g"), "");
            finalXmlContent = finalXmlContent.replace(new RegExp(`</${xmlElement}>`, "g"), "");
            finalXmlContent = finalXmlContent.replace(/<!\[CDATA\[\s*([\s\S]*?)\s*]]>/g, function (_, cdataContent) {
                const trimmedContent = cdataContent.trim();
                const lines = trimmedContent.split("\n");
                const indentedLines = lines.map(function (line) {
                    return line.replace(/^\s*/, "");
                });
                return ("<![CDATA[\n" + INDENT + indentedLines.join("\n" + INDENT) + "\n]]>");
            });
            finalXmlContent = finalXmlContent.replace(/(\n\s*){2,}/g, `\n${INDENT}`);
            const closeTag = `</${xmlElement}>`;
            yield promises__namespace.writeFile(filePath, `${XML_HEADER}\n${xmlRootElementHeader}${finalXmlContent}${closeTag}`);
            logger.debug(`Created reassembled file: ${filePath}`);
        });
    }

    function buildRootElementHeader(rootElement, rootElementName) {
        let rootElementHeader = `<${rootElementName}`;
        for (const [attrKey, attrValue] of Object.entries(rootElement)) {
            if (attrKey.startsWith("@")) {
                const cleanAttrKey = attrKey.slice(2);
                rootElementHeader += ` ${cleanAttrKey}="${String(attrValue)}"`;
            }
        }
        rootElementHeader += ">";
        return rootElementHeader;
    }

    function buildXMLString(element, indentLevel = 0) {
        const xmlBuilder = new fastXmlParser.XMLBuilder(JSON_PARSER_OPTION);
        const xmlString = xmlBuilder.build(element);
        const formattedXml = xmlString
            .split("\n")
            .map((line) => `${" ".repeat(indentLevel * INDENT.length)}${line}`)
            .join("\n")
            .trimEnd();
        return formattedXml;
    }

    const xmlParser = new fastXmlParser.XMLParser(XML_PARSER_OPTION);
    class ReassembleXMLFileHandler {
        processFilesInDirectory(dirPath) {
            return __awaiter(this, void 0, void 0, function* () {
                const combinedXmlContents = [];
                let rootResult = undefined;
                const files = yield promises__namespace.readdir(dirPath);
                files.sort((fileA, fileB) => {
                    const fullNameA = fileA.split(".")[0].toLowerCase();
                    const fullNameB = fileB.split(".")[0].toLowerCase();
                    return fullNameA.localeCompare(fullNameB);
                });
                for (const file of files) {
                    const filePath = path__namespace.join(dirPath, file);
                    const fileStat = yield promises__namespace.stat(filePath);
                    if (fileStat.isFile() && filePath.endsWith(".xml")) {
                        const xmlContent = yield promises__namespace.readFile(filePath, "utf-8");
                        let xmlParsed;
                        try {
                            xmlParsed = xmlParser.parse(xmlContent, true);
                        }
                        catch (err) {
                            logger.error(`${filePath} was unable to be parsed and was not added to the reassembled file. Confirm formatting and try again.`);
                            continue;
                        }
                        const rootResultFromFile = yield this.processFilesForRootElement(xmlParsed);
                        if (rootResultFromFile && !rootResult) {
                            rootResult = rootResultFromFile;
                        }
                        const combinedXmlString = buildXMLString(xmlParsed);
                        combinedXmlContents.push(combinedXmlString);
                    }
                    else if (fileStat.isDirectory()) {
                        const [subCombinedXmlContents, subRootResult] = yield this.processFilesInDirectory(filePath);
                        combinedXmlContents.push(...subCombinedXmlContents);
                        if (subRootResult && !rootResult) {
                            rootResult = subRootResult;
                        }
                    }
                }
                return [combinedXmlContents, rootResult];
            });
        }
        processFilesForRootElement(xmlParsed) {
            return __awaiter(this, void 0, void 0, function* () {
                const rootElementName = Object.keys(xmlParsed)[1];
                const rootElement = xmlParsed[rootElementName];
                const rootElementHeader = buildRootElementHeader(rootElement, rootElementName);
                return [rootElementName, rootElementHeader];
            });
        }
        reassemble(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, fileExtension, postPurge = false } = xmlAttributes;
                let combinedXmlContents = [];
                const fileStat = yield promises__namespace.stat(xmlPath);
                if (!fileStat.isDirectory()) {
                    logger.error(`The provided xmlPath ${xmlPath} to reassemble is not a directory.`);
                    return;
                }
                logger.debug(`Parsing directory to reassemble: ${xmlPath}`);
                const [subCombinedXmlContents, rootResult] = yield this.processFilesInDirectory(xmlPath);
                combinedXmlContents = subCombinedXmlContents;
                const parentDirectory = path__namespace.dirname(xmlPath);
                const subdirectoryBasename = path__namespace.basename(xmlPath);
                const fileName = fileExtension
                    ? `${subdirectoryBasename}.${fileExtension}`
                    : `${subdirectoryBasename}.xml`;
                const filePath = path__namespace.join(parentDirectory, fileName);
                if (rootResult !== undefined) {
                    const [rootElementName, rootElementHeader] = rootResult;
                    yield buildReassembledFile(combinedXmlContents, filePath, rootElementName, rootElementHeader);
                    if (postPurge)
                        yield promises__namespace.rm(xmlPath, { recursive: true });
                }
                else {
                    logger.error(`No files under ${xmlPath} were parsed successfully. A reassembled XML file was not created.`);
                }
            });
        }
    }

    function findUniqueIdElement(element, uniqueIdElements) {
        if (uniqueIdElements === undefined) {
            return getShortHash(element);
        }
        const uniqueIdElementsArray = uniqueIdElements.split(",");
        for (const fieldName of uniqueIdElementsArray) {
            if (element[fieldName] !== undefined) {
                if (typeof element[fieldName] === "string") {
                    return element[fieldName];
                }
            }
        }
        for (const key in element) {
            if (typeof element[key] === "object" && element[key] !== null) {
                const childFieldName = findUniqueIdElement(element[key], uniqueIdElements);
                if (childFieldName !== undefined) {
                    return childFieldName;
                }
            }
        }
        return getShortHash(element);
    }
    function getShortHash(element) {
        const hash = node_crypto.createHash("sha256");
        hash.update(JSON.stringify(element));
        const fullHash = hash.digest("hex");
        return fullHash.slice(0, 8);
    }

    function buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, parentKey, indent) {
        return __awaiter(this, void 0, void 0, function* () {
            let elementContent = "";
            const fieldName = findUniqueIdElement(element, uniqueIdElements);
            const outputDirectory = path__namespace.join(metadataPath, parentKey);
            const outputFileName = `${fieldName}.${parentKey}-meta.xml`;
            const outputPath = path__namespace.join(outputDirectory, outputFileName);
            yield promises__namespace.mkdir(outputDirectory, { recursive: true });
            const parentKeyHeader = buildRootElementHeader(element, parentKey);
            elementContent = buildXMLString(element, 2);
            let decomposeFileContents = `${XML_HEADER}\n`;
            decomposeFileContents += `${rootElementHeader}\n`;
            decomposeFileContents += `${indent}${parentKeyHeader}\n`;
            decomposeFileContents += `${elementContent}\n`;
            decomposeFileContents += `${indent}</${parentKey}>\n`;
            decomposeFileContents += `</${rootElementName}>`;
            yield promises__namespace.writeFile(outputPath, decomposeFileContents);
            logger.debug(`Created disassembled file: ${outputPath}`);
        });
    }

    function processElement(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const { element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, key, indent, leafContent, leafCount, hasNestedElements, } = params;
            if (typeof element === "object") {
                yield buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, key, indent);
                return [leafContent, leafCount, true];
            }
            else {
                const fieldValue = element;
                const updatedLeafContent = `${leafContent}${indent}<${key}>${String(fieldValue)}</${key}>\n`;
                return [updatedLeafContent, leafCount + 1, hasNestedElements];
            }
        });
    }

    function buildLeafFile(leafContent, metadataPath, baseName, rootElementName, rootElementHeader) {
        return __awaiter(this, void 0, void 0, function* () {
            let leafFile = `${XML_HEADER}\n`;
            leafFile += `${rootElementHeader}\n`;
            const sortedLeafContent = leafContent
                .split("\n")
                .filter((line) => line.trim() !== "")
                .sort((a, b) => a.localeCompare(b))
                .join("\n");
            leafFile += sortedLeafContent;
            leafFile += `\n</${rootElementName}>`;
            const leafOutputPath = path__namespace.join(metadataPath, `${baseName}.xml`);
            yield promises__namespace.writeFile(leafOutputPath, leafFile);
            logger.debug(`Created disassembled file: ${leafOutputPath}`);
        });
    }

    function buildDisassembledFiles(xmlString, metadataPath, uniqueIdElements, baseName, indent, postPurge, parentPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const xmlParser = new fastXmlParser.XMLParser(XML_PARSER_OPTION);
            let result;
            try {
                result = xmlParser.parse(xmlString, true);
            }
            catch (err) {
                logger.error(`${baseName}.xml was unable to be parsed. Confirm formatting and try again.`);
                return;
            }
            const rootElementName = Object.keys(result)[1];
            const rootElement = result[rootElementName];
            const rootElementHeader = buildRootElementHeader(rootElement, rootElementName);
            let leafContent = "";
            let leafCount = 0;
            let hasNestedElements = false;
            for (const key of Object.keys(rootElement).filter((key) => !key.startsWith("@"))) {
                if (Array.isArray(rootElement[key])) {
                    for (const element of rootElement[key]) {
                        const [updatedLeafContent, updatedLeafCount, updatedHasNestedElements] = yield processElement({
                            element,
                            metadataPath,
                            uniqueIdElements,
                            rootElementName,
                            rootElementHeader,
                            key,
                            indent,
                            leafContent,
                            leafCount,
                            hasNestedElements,
                        });
                        leafContent = updatedLeafContent;
                        leafCount = updatedLeafCount;
                        hasNestedElements = updatedHasNestedElements;
                    }
                }
                else {
                    const [updatedLeafContent, updatedLeafCount, updatedHasNestedElements] = yield processElement({
                        element: rootElement[key],
                        metadataPath,
                        uniqueIdElements,
                        rootElementName,
                        rootElementHeader,
                        key,
                        indent,
                        leafContent,
                        leafCount,
                        hasNestedElements,
                    });
                    leafContent = updatedLeafContent;
                    leafCount = updatedLeafCount;
                    hasNestedElements = updatedHasNestedElements;
                }
            }
            if (!hasNestedElements) {
                logger.error(`The XML file ${baseName}.xml only has leaf elements. This file will not be disassembled.`);
                return;
            }
            if (leafCount > 0) {
                yield buildLeafFile(leafContent, metadataPath, baseName, rootElementName, rootElementHeader);
            }
            if (postPurge) {
                const originalFilePath = path__namespace.resolve(`${parentPath}/${baseName}.xml`);
                promises__namespace.unlink(originalFilePath);
            }
        });
    }

    class DisassembleXMLFileHandler {
        disassemble(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, uniqueIdElements, prePurge = false, postPurge = false, } = xmlAttributes;
                const fileStat = yield promises__namespace.stat(xmlPath);
                if (fileStat.isFile()) {
                    const filePath = path__namespace.resolve(xmlPath);
                    if (!filePath.endsWith(".xml")) {
                        logger.error(`The file path ${filePath} is not an XML file.`);
                        return;
                    }
                    const basePath = path__namespace.dirname(filePath);
                    yield this.processFile({
                        xmlPath: basePath,
                        filePath,
                        uniqueIdElements,
                        prePurge,
                        postPurge,
                    });
                }
                else if (fileStat.isDirectory()) {
                    const files = yield promises__namespace.readdir(xmlPath);
                    for (const file of files) {
                        const filePath = path__namespace.join(xmlPath, file);
                        if (filePath.endsWith(".xml")) {
                            yield this.processFile({
                                xmlPath,
                                filePath,
                                uniqueIdElements,
                                prePurge,
                                postPurge,
                            });
                        }
                    }
                }
            });
        }
        processFile(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, filePath, uniqueIdElements, prePurge, postPurge } = xmlAttributes;
                logger.debug(`Parsing file to disassemble: ${filePath}`);
                const xmlContent = yield promises__namespace.readFile(filePath, "utf-8");
                const fullName = path__namespace.basename(filePath, path__namespace.extname(filePath));
                const baseName = fullName.split(".")[0];
                let outputPath;
                outputPath = path__namespace.join(xmlPath, baseName);
                if (prePurge) {
                    if (fs__namespace.existsSync(outputPath)) {
                        yield promises__namespace.rm(outputPath, { recursive: true });
                    }
                }
                yield buildDisassembledFiles(xmlContent, outputPath, uniqueIdElements, fullName, INDENT, postPurge, xmlPath);
            });
        }
    }

    function setLogLevel(level) {
        log4js__namespace.getLogger().level = level;
    }
    const logger = log4js__namespace.getLogger();
    log4js__namespace.configure({
        appenders: { disassemble: { type: "file", filename: "disassemble.log" } },
        categories: { default: { appenders: ["disassemble"], level: "error" } },
    });

    exports.DisassembleXMLFileHandler = DisassembleXMLFileHandler;
    exports.ReassembleXMLFileHandler = ReassembleXMLFileHandler;
    exports.logger = logger;
    exports.setLogLevel = setLogLevel;

}));
//# sourceMappingURL=index.umd.js.map
